from django.core.exceptions import ValidationError

from fabric.utils import abort, _AttributeDict


class Settings(_AttributeDict):
    """
    Object that holds setting items and a method of validating required
    settings.

    The class should be subclassed and the key `defaults` should contain a dict
    with default values.

    The dict `required` should contain key => (`validator`, ...) where
    `validator` is a callable that will be called with the value to be
    validated.
    """
    def __init__(self, **kwargs):
        super(Settings, self).__init__()

        self.update(self.defaults)
        self.update(kwargs)

    def resolve_all_keys(self):
        def replace(v):
            """
            Recursivly replace strings, dicts, tuples and lists.
            """
            if isinstance(v, (tuple, list)):
                return map(replace, v)
            elif isinstance(v, dict):
                for key, value in v.iteritems():
                    v[replace(key)] = replace(value)
                return v
            elif isinstance(v, basestring):
                return v % self
            else:
                return v

        for key, value in self.iteritems():
            self[key] = replace(value)

    def validate_or_abort(self):
        """
        Validate the values acording to the list of validators given.

        Also resolves all values if validation was successful.
        """
        missing = []

        for key, validators in self.required.iteritems():
            if not isinstance(validators, (list, tuple)):
                validators = (validators, )

            for validator in validators:
                try:
                    validator(self[key])
                except KeyError:
                    missing.append("* %s was not defined" % key)
                except ValueError, exc:
                    missing.append("* %s caused ValueError(%s)" % (key, exc))
                except ValidationError, exc:
                    missing.append("* %s failed validation - %s" % (
                        key, "\n  * ".join(exc.messages)
                    ))

        if len(missing) > 0:
            abort(
                "missing (or invalid) settings:\n\n" + ("\n".join(missing))
            )

        self.resolve_all_keys()



from django.core.validators import (
    MinLengthValidator, RegexValidator
)

from coat.settings import Settings


__all__ = ("DjangoSettings", "VirtualEnvSettings")


class VirtualEnvSettings(Settings):
    """
    A settings object for a python virtualenv based envorinment.
    """
    DEFAULT_ACTIVATOR = 'source {dir}/bin/activate'

    DEFAULT_COMMANDS = [
        "pip -q install -r django/requirements.txt"
    ]

    DEFAULT_INIT_COMMANDS = [
        "virtualenv %(env_dir)s"
    ]

    defaults = {
        'activator': DEFAULT_ACTIVATOR,
        'commands': DEFAULT_COMMANDS,
        'init_commands': DEFAULT_INIT_COMMANDS,
    }

    required = {
        'env_dir': MinLengthValidator(1),
    }


class DjangoSettings(Settings):
    """
    A settings object for a Django project.
    """
    DEFAULT_MANAGEMENT_COMMANDS = [
        "syncdb --noinput",
    ]

    defaults = {
        'management_commands': DEFAULT_MANAGEMENT_COMMANDS,
        'versions_dir': 'app/versions',
    }

    required = {
        'settings_file': RegexValidator(
            "^localsettings_(\w+)\.py$",
            "localsettings file must be defined and named localsettings_ENV.py"
        ),
    }



from __future__ import with_statement

import os
import re
import tempfile
import shutil

from datetime import datetime

from fabric.api import run, local, get, cd, lcd, put
from fabric.state import env
from fabric.operations import require
from fabric.contrib.console import confirm

from .base import get_local_base_dir, backup_create_local


__all__ = ("update_env", "deploy")


def update_env(*args, **kwargs):
    env.local_path = "public_html"
    env.update(kwargs)

    if 'local_base_dir' not in env:
        env.local_base_dir = get_local_base_dir()


def deploy():
    require('base_dir', provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    deploy_archive_dir = tempfile.mkdtemp()

    with lcd(env.local_base_dir):
        local('git archive master %s | tar -x -f- -C %s' % (env.local_path, deploy_archive_dir))

    if 'settings_file' in env:
        command = 'rsync -a --exclude config.php  %s/%s/* %s@%s:%s/%s'
    else:
        command = 'rsync -a %s/%s/* %s@%s:%s/%s'

    local(command %
          (deploy_archive_dir, env.local_path, env.user, env.host,
           env.base_dir, env.remote_path))

    with cd(os.path.join(env.base_dir, env.remote_path)):
        if env.settings_file:
            run('cp %s config.php' % env.settings_file)

    shutil.rmtree(deploy_archive_dir)



from __future__ import with_statement

import os
import shutil
import glob

from pydispatch import dispatcher
from fabric.api import run, local, cd, settings, prefix, hide
from fabric.operations import require
from fabric.state import env
from fabric.contrib import files as fabric_files

from coat import utils as coat_utils
from coat.django.utils import find_settings, find_manage
from coat.django import signals


def copy_revision_to_remote(workdir, remote_revision, deploy_revision):
    remote_versions_dir = coat_utils.remote_absolute_path(
        env.django_settings.versions_dir
    )

    if not fabric_files.exists(remote_versions_dir):
        run("mkdir -p %s" % remote_versions_dir)

    remote_deploy_dir = "%s/%s" % (remote_versions_dir, deploy_revision)

    # test remote for the revision and skip re-copying already
    # existing revisions
    if fabric_files.exists("%s" % remote_deploy_dir):
        return

    # copy all of local onto remote using rsync, use hard links to save
    # space for unmodified files
    rsync_cmd = ("%s/* %s@%%s:%s/" %
        (workdir, env.user, remote_deploy_dir)
    )

    dispatcher.send(
        signal=signals.pre_workdir_copy_to_remote,
        sender=copy_revision_to_remote,
        deploy_revision=deploy_revision,
        remote_revision=remote_revision,
        workdir=workdir,
    )

    if remote_revision:
        with cd("%s" % remote_versions_dir):
            run("rsync -a --link-dest=../%(cur)s/ %(cur)s/ %(new)s" %
                {'cur': remote_revision, 'new': deploy_revision})

        rsync_cmd = "-a -c --delete " + rsync_cmd
    else:
        rsync_cmd = "-a " + rsync_cmd

    for host in env.hosts:
        port = env.port

        if ":" in host:
            host, port = host.rsplit(":", 1)

        rsync_cmd += " --rsh='ssh -p%s'" % port

        local("rsync %s" % (rsync_cmd % host))

    dispatcher.send(
        signal=signals.post_workdir_copy_to_remote,
        sender=copy_revision_to_remote,
        deploy_revision=deploy_revision,
        remote_revision=remote_revision,
        workdir=workdir,
    )


def remote_activate_revision(workdir, remote_revision, deploy_revision):
    remote_virtualenv_dir = "%s/%s" % (
        env.base_dir, env.virtualenv_settings.env_dir
    )

    remote_versions_dir = "%s/%s" % (
        env.base_dir, env.django_settings.versions_dir
    )

    dispatcher.send(
        signal=signals.pre_remote_run_commands,
        sender=remote_activate_revision,
    )

    # initialize virtualenv if non-existant
    if not fabric_files.exists(remote_virtualenv_dir):
        dispatcher.send(
            signal=signals.pre_remote_init_virtualenv,
            sender=remote_activate_revision
        )

        with cd(env.base_dir):
            for command in env.virtualenv_settings.init_commands:
                run(command)

        dispatcher.send(
            signal=signals.post_remote_init_virtualenv,
            sender=remote_activate_revision,
        )

    # find relative path to manage.py
    django_manage_path = find_manage(workdir).replace(workdir, "")[1:]

    with cd("%s/%s" % (remote_versions_dir, deploy_revision)):
        with prefix(env.virtualenv_settings.activator.format(dir=remote_virtualenv_dir)):
            for command in env.virtualenv_settings.commands:
                run(command)

        for command in env.django_settings.management_commands:
            run("%s/bin/python %s %s" % (
                remote_virtualenv_dir, django_manage_path, command
            ))

    dispatcher.send(
        signal=signals.post_remote_run_commands,
        sender=remote_activate_revision,
    )

    dispatcher.send(
        signal=signals.pre_remote_activate_revision,
        sender=remote_activate_revision,
    )

    with cd(remote_versions_dir):
        with settings(hide("warnings", "stderr"), warn_only=True):
            run("test -L current && rm current")

        run("ln -s %s current" % deploy_revision)

    dispatcher.send(
        signal=signals.post_remote_activate_revision,
        sender=remote_activate_revision,
    )


def remote_reload():
    dispatcher.send(
        signal=signals.pre_remote_reload,
        sender=remote_reload,
    )

    if getattr(env.django_settings, 'wsgi_file', False):
        run("touch %s" %
            coat_utils.remote_absolute_path(env.django_settings.wsgi_file)
            )

    dispatcher.send(
        signal=signals.post_remote_reload,
        sender=remote_reload,
    )


def workdir_django_prepare(workdir):
    if 'settings_file' in env.django_settings:
        django_basedir = os.path.dirname(find_settings(workdir))

        shutil.copy(
            os.path.join(django_basedir, env.django_settings.settings_file),
            os.path.join(django_basedir, "localsettings.py")
        )

        map(os.unlink, glob.glob("%s/localsettings_*.py" % django_basedir))


def deploy(revision="master"):
    require("django_settings", "base_dir", "virtualenv_settings",
            provided_by=("env_test", "env_live"))

    env.remote_pwd = run("pwd")

    env.django_settings.validate_or_abort()
    env.virtualenv_settings.validate_or_abort()

    env.remote_revision = coat_utils.remote_resolve_current_revision()
    env.deploy_revision = coat_utils.local_resolve_revision(revision)

    env.deploy_workdir = coat_utils.workdir_prepare_checkout(
        revision, folders=("django", )
    )

    workdir_django_prepare(env.deploy_workdir)

    copy_revision_to_remote(
        env.deploy_workdir, env.remote_revision, env.deploy_revision
    )

    remote_activate_revision(
        env.deploy_workdir, env.remote_revision, env.deploy_revision
    )

    remote_reload()



from __future__ import with_statement

import os
import re
import tempfile
import time
import shutil

from datetime import datetime

from fabric.api import run, local, get, cd, lcd, put
from fabric.state import env
from fabric.operations import require
from fabric.contrib.console import confirm

__all__ = ("update_env", "download_uploads_from_remote",
           "upload_uploads_to_remote", "download_database_from_remote",
           "update_database_from_remote", "update_database_on_remote",
           "deploy")


def get_local_base_dir():
    return os.path.dirname(os.path.abspath(env.real_fabfile))


def backup_prune():
    if not env.backup_path:
        return

    backup_path = os.path.join(env.local_base_dir, env.backup_path)

    if not os.path.exists(backup_path):
        return

    wanted = []

    for entry in os.listdir(backup_path):
        if entry.startswith("."):
            continue

        absentry = os.path.join(backup_path, entry)

        if time.time() - os.path.getctime(absentry) > 43200:
            wanted.append(absentry)

    for entry in wanted:
        shutil.rmtree(entry)


def backup_create_local(local_source):
    backup_prune()

    backup_target = os.path.join(env.local_base_dir,
                                 env.backup_path,
                                 env.backup_prefix)

    if os.path.isdir(local_source):
        backup_target = os.path.join(backup_target, local_source)

    local("mkdir -p %s" % backup_target)

    local("rsync -a --link-dest=%s %s %s" % (
        local_source, local_source, backup_target)
    )


def update_env(*args, **kwargs):
    env.wordpress_path = "public_html"
    env.backup_prefix = "%Y%m%d-%H%m%S"
    env.backup_path = "backups"

    env.update(kwargs)

    if 'local_base_dir' not in env:
        env.local_base_dir = get_local_base_dir()

    if env.backup_prefix:
        env.backup_prefix = datetime.now().strftime(env.backup_prefix)


def parse_config_from_file(settings_file):
    lines = open(settings_file).read()

    define_re = re.compile(r"""define\(["'](?P<key>[^'"]+)["'],\s*["'](?P<value>[^'"]*)["']\)""", re.S | re.M)

    return dict(define_re.findall(lines))


def read_config():
    """
    Returns a dict of the current local and remote wordpress settings.
    """
    wordpress_path = os.path.join(env.local_base_dir, env.local_wordpress_path)

    return {
        'local': parse_config_from_file(os.path.join(wordpress_path, "wp-config.php")),
        'remote': parse_config_from_file(os.path.join(wordpress_path, env.settings_file))
    }


def download_uploads_from_remote():
    require("base_dir", provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    with lcd(os.path.join(env.local_base_dir, env.local_wordpress_path)):
        backup_create_local("wp-content/uploads")

        local("rsync -a "
              "%(user)s@%(host)s:%(base_dir)s/%(wordpress_path)s/wp-content/uploads/ "
              "wp-content/uploads/" % env)


def upload_uploads_to_remote():
    require("base_dir", provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    confirm("This will override data on remote. Are you sure?", default=False)

    with lcd(os.path.join(env.local_base_dir, env.local_wordpress_path)):
        local("rsync -a wp-content/uploads/* "
              "%(user)s@%(host)s:%(base_dir)s/%(wordpress_path)s/wp-content/uploads/" % env)


def download_database_from_remote():
    require("base_dir", provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    wp_config = read_config()

    run("mysqldump -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s --add-drop-table %(DB_NAME)s > /tmp/%(DB_USER)s.sql" % wp_config['remote'])
    get("/tmp/%(DB_USER)s.sql" % wp_config['remote'], os.path.join(env.local_base_path, env.local_wordpress_path, "%(DB_USER)s.sql" % wp_config['remote']))
    run("rm -f /tmp/%(DB_USER)s.sql" % wp_config['remote'])


def update_database_from_remote():
    require("base_dir", provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    wp_config = read_config()
    dump_filename = "/tmp/%(DB_USER)s_%(DB_NAME)s.sql" % wp_config['remote']

    wp_config['remote']['dump_filename'] = dump_filename
    wp_config['local']['dump_filename'] = dump_filename

    with lcd(env.local_base_dir):
        local("mysqldump -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s --add-drop-table %(DB_NAME)s | gzip > %(DB_NAME)s.sql.gz" % wp_config['local'])

        backup_create_local("%(DB_NAME)s.sql.gz" % wp_config['local'])

        run("rm -f %(DB_NAME)s.sql.gz" % wp_config['local'])

    run("mysqldump -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s --add-drop-table %(DB_NAME)s > %(dump_filename)s" % wp_config['remote'])
    get(dump_filename, dump_filename)
    run("rm -f %s" % dump_filename)

    local("mysql -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s %(DB_NAME)s < %(dump_filename)s" % wp_config['local'])

    os.unlink(dump_filename)


def update_database_on_remote():
    require("base_dir", provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    confirm("This will override data on remote. Are you sure?", default=False)

    wp_config = read_config()

    local("mysqldump -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s --add-drop-table %(DB_NAME)s > /tmp/%(DB_USER)s.sql" % wp_config['local'])
    put("/tmp/%(DB_USER)s.sql" % wp_config['local'], "/tmp/%(DB_USER)s.sql" % wp_config['remote'])
    os.unlink("/tmp/%(DB_USER)s.sql" % wp_config['local'])

    wp_config['remote']['backup_name'] = "%(DB_USER)s_%(DB_NAME)s" % wp_config['remote']

    run("mysqldump -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s --add-drop-table %(DB_NAME)s | gzip > /tmp/%(backup_name)s.sql.gz" % wp_config['remote'])

    run("mysql -u%(DB_USER)s -p\"%(DB_PASSWORD)s\" -h%(DB_HOST)s %(DB_NAME)s < /tmp/%(DB_USER)s.sql" % wp_config['remote'])
    run("rm -f /tmp/%(DB_USER)s.sql" % wp_config['remote'])


def deploy(revision="master", preprocessor=None):
    require('base_dir', provided_by=("env_test", "env_live"),
            used_for='defining the deploy environment')

    deploy_archive_dir = tempfile.mkdtemp()

    with lcd(env.local_base_dir):
        local('git archive %s %s | tar -x -f- -C %s' % (revision, env.local_wordpress_path, deploy_archive_dir))

        if preprocessor:
            preprocessor()

    local('rsync -a --exclude wp-config.php --exclude wp-content/uploads/* %s/%s/* %s@%s:%s/%s' %
          (deploy_archive_dir, env.local_wordpress_path, env.user, env.host,
           env.base_dir, env.wordpress_path))

    with cd(os.path.join(env.base_dir, env.wordpress_path)):
        if env.settings_file:
            run('cp %s wp-config.php' % env.settings_file)

    shutil.rmtree(deploy_archive_dir)



#coding: utf-8
import os
from setuptools import setup, find_packages


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(
    name="fabric-coat",
    version="0.4.9",
    author="Rasmus Schlünsen, Mads Sülau Jørgensen",
    author_email="rs@konstellation.dk, msj@konstellation.dk",
    description=("Deployment helpers for fabric"),
    license="BSD",
    keywords="fabric coat deployment rsync helper",
    url="https://bitbucket.org/madssj/fabric-coat",
    packages=find_packages('src'),
    long_description=read('README'),
    package_dir={'': 'src'},
    install_requires=['fabric >= 1.3', 'pydispatcher'],
)



import os
import atexit
import shutil
import tempfile

from pydispatch import dispatcher
from fabric.state import env
from fabric.api import lcd, local, hide, settings, cd, run
from fabric.contrib import files as fabric_files
from coat import signals


def get_project_root_directory():
    """
    Returns the dirname of the fabfile.
    """
    return os.path.dirname(os.path.abspath(env.real_fabfile))


def workdir_prepare_checkout(revision, folders):
    """
    Returns an absolute path to a directory containing a git checkout of the
    given treeish revision.

    Registers a hook with atexit to delete the returned directory.

    Dispatches signals to pre_workdir_checkout and post_workdir_checkout.
    """
    dispatcher.send(
        signal=signals.pre_workdir_prepare_checkout,
        sender=workdir_prepare_checkout,
        revision=revision,
    )

    workdir = tempfile.mkdtemp()

    atexit.register(shutil.rmtree, workdir, True)

    with lcd(get_project_root_directory()):
        local("git archive %s %s | tar -x -C %s" %
              (revision, " ".join(folders), workdir))

        dispatcher.send(
            signal=signals.post_workdir_prepare_checkout,
            sender=workdir_prepare_checkout,
            revision=revision,
            workdir=workdir,
        )

    return workdir


def local_resolve_revision(revision):
    """
    Returns a partial sha1 from a treeish revision.
    """
    with lcd(get_project_root_directory()):
        with hide("running"):
            revision = local("git log -1 --format=%%h %s" % revision,
                             capture=True)

    return revision


def remote_absolute_path(path):
    """
    Returns an absolute path built of joining remote_pwd and base_dir with
    path. If either is an absolute path, it will be used returned.

    Must be used everywhere a remote path is used.
    """
    return os.path.join(env.remote_pwd, env.base_dir, path)


def remote_resolve_current_revision():
    """
    Returns the locally resolved currently active remote revision.
    """
    revision = None

    with settings(hide("stderr", "running", "stdout", "warnings"),
                  warn_only=True):

        # make sure version_dir exists
        versions_dir = remote_absolute_path(env.django_settings.versions_dir)

        if not fabric_files.exists(versions_dir):
            run("mkdir -p %s" % versions_dir)

        # try to resolve current symlink
        with cd(versions_dir):
            if fabric_files.exists("current"):
                revision = run("readlink current")

                if not revision.failed:
                    revision = revision.rstrip("/")

    if revision:
        return local_resolve_revision(revision)
